// Hardcore Compo Virtual Machine
// See http://www.uw-team.org/forum/viewtopic.php?p=51972
// 
// Code by gynvael.coldwind//vx
// http://gynvael.vexillium.org
// http://vexillium.org
//
// BSD-Style License...
//
//
//  TODO:
//
//
//
//  CHANGE LOG:
//
// beta2.1->beta2.1b:
//  * Minor changes - just added benchmark option that shows
//    current CPU cycles per second. Compile with BENCHMARK
//    preprocessor option.
//
//
// beta2->beta2.1:
//  * User-Opcodes can be used while Interrupt Block is set, but 
//    Special.IntReturnAddr is still overwritten (but since this is
//    deterministic, it should not cause any problems).
//  * Special.IntBlocker is no longer a flag, but a block counter. 
//    If it is equal to 0, no block is present. If it is larger then
//    0, interrupt block is present. Every entry to an interupt block
//    increments the value (BlockInt opcode too). Every ReleaseInt
//    decreases the value (after one opcode delay ofc). This changes
//    almost nothing, but allows User-Opcodes to be called when
//    interrupts are blocked without any changes in the code.
//
//
// beta1->beta2:
//  * Added Special.CursorHidden (02005004) flag
//  * ReleaseInt opcode (0x15) now delays the release one instruction
//    This allows to create interrupt handling like this:
//    int_handler:
//      ...
//      ReleaseInt
//      Goto [Special.IntReturnAddr]
//    The programmer does not have to worry about Special.IntReturnAddr being
//    overwritten in this case, since ReleaseInt will unblock interrupts after
//    Goto.
//    Please note that there must be one instruction delay after ReleaseInt in
//    using custom opcodes and exception handling.
//  * Added Middle Mouse Button support. State can be read at 
//    Special.MouseMMBState (02003018)
//  * Fixed invalid Goto/Special.IP write VM DoS. Now it issues an Invalid 
//    Address exception.
//  * Fixed mouse position race condition. Now mouse position is NOT updated if
//    Special.IntBlocker is set. However Mouse button state is updated.
//  * Added Interrupt Queue. Timer1/Timer2/Mouse/Keyboard events are queued if
//    they cannot be handled right away. The queue size is 128 events.
//  * Added flag that disables queueing mouse move events 
//    (Specia.DisableMouseMoveQueue at 02000018). Disabling mouse move events
//    tweaks up performance, but some move events are lost.
//  * Added SIMD (Single Instruction Multiple Data) opcodes:
//    0x16 - 0x1E - SESESE group
//    0x1F - 0x23 - SESESE2 group
// 
//
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <SDL/SDL.h>
#include "vm.h"

#undef main

InstructionFunc FuncPtrs[256] = {
  Halt,
  Move_MemImm,
  Move_MemMem,
  Add_MemImm,
  Add_MemMem,
  And_MemImm,
  And_MemMem,
  Or_MemImm,
  Or_MemMem,
  Xor_MemImm,
  Xor_MemMem,
  Shl_MemImm,
  Shr_MemImm,
  Goto_Mem,
  Goto_Imm,
  Neg_Mem,
  MovePtrFrom_MemMem,
  MovePtrTo_MemMem,
  GotoEqual_ImmMemMem,
  GotoLess_ImmMemMem,
  BlockInt,
  ReleaseInt,
  SIMDMove_MemMemMem, // SESESE
  SIMDAdd_MemMemMem,
  SIMDAnd_MemMemMem,
  SIMDOr_MemMemMem,
  SIMDXor_MemMemMem,
  SIMDShl_MemImmMem,
  SIMDShr_MemImmMem,
  SIMDNeg_MemMem,
  SIMDCopy_MemMemMem,
  SIMDAdd_MemImmMem, // SESESE2
  SIMDAnd_MemImmMem,
  SIMDOr_MemImmMem,
  SIMDXor_MemImmMem,
  SIMDSet_MemImmMem,
};

unsigned char Rom[0x1000000];
unsigned char Ram[0x1000000];

union
{
  unsigned char Data[0x1000000];
  struct
  {
    // --------------------------------------------------------------
    // General
    // --------------------------------------------------------------

    // 02000000    4    Current Instruction Pointer
    unsigned int IP;

    // 02000004    4    Interrupt Return Addr
    unsigned int IntReturnAddr;

    // 02000008    4    Inside interrupt/exception/user opcode flag (aka Int Blocker)
    unsigned int IntBlocker;

    // 0200000C    4    Time (ms from execution)
    unsigned int TimeMS;

    // 02000010    4    Executed Instruction Counter LOW
    unsigned int ExecInstCounterLow;

    // 02000014    4    Executed Instruction Counter HIGH
    unsigned int ExecInstCounterHi;

    // 02000018    4    Disable Mouse Move Event Queue
    unsigned int DisableMouseMoveQueue;

    // 0200001C         Padding to 02001000
    unsigned char Reserved1[0x1000 - 7 * 4]; // XXX Keep this up to date!!!

    // --------------------------------------------------------------
    // Exceptions
    // --------------------------------------------------------------

    // 02001000    4    Invalid Memory Address Exception Handler Vector
    unsigned int ExceptionInvAddrHandler;

    // 02001004    4    ROM Write Exception Handler Vector
    unsigned int ExceptionROMWriteErrorHandler;

    // 02001008    4    Invalid Opcode Exception Handler Vector
    unsigned int ExceptionInvalidOpcodeHandler;

    // 0200100C    4    Timer Interrupt Frequency (in ms)
    unsigned int Timer1IntFreq;

    // 02001010    4    Timer Interrupt Handler
    unsigned int Timer1Handler;

    // 02001014    4    Opcode Timer Interrupt Frequency (in opcodes)
    unsigned int Timer2IntFreq;

    // 02001018    4    Opcode Timer Interrupt Handler
    unsigned int Timer2Handler;

    // 0200101C         Padding to 02002000
    unsigned char Reserved2[0x1000 - 7 * 4];  // XXX Keep this up to date!!!

    // --------------------------------------------------------------
    // Keyboard
    // --------------------------------------------------------------

    // 02002000    4    Key Up Handler
    unsigned int KeyUpHandler;

    // 02002004    4    Key Down Handler
    unsigned int KeyDownHandler;

    // 02002008    4    Key Handled
    unsigned int KeyHandled;

    // 0200200C    4    Padding
    unsigned int PaddingReserved;

    // 02002010  SDLK_LAST * 4  Keyboard Map
    unsigned int KeyMap[SDLK_LAST];

    // 02002???         Padding to 02003000
    unsigned char Reserved3[0x1000 - SDLK_LAST * 4 - 4 * 4];

    // --------------------------------------------------------------
    // Mouse
    // --------------------------------------------------------------

    // 02003000    4    Mouse X
    unsigned int MouseX;

    // 02003004    4    Mouse Y
    unsigned int MouseY;

    // 02003008    4    Mouse Left Mouse Button State
    unsigned int MouseLMBState;

    // 0200300C    4    Mouse Right Mouse Button State
    unsigned int MouseRMBState;

    // 02003010    4    Mouse Move Handler
    unsigned int MouseMoveHandler;

    // 02003014    4    Mouse Button Change State Handler
    unsigned int MouseButtonChangeStateHandler;

    // 02003018    4    Mouse Middle Mouse Button State
    unsigned int MouseMMBState;

    // 0200301C         Padding to 02004000
    unsigned char Reserved4[0x1000 - 7 * 4]; // XXX Keep this up to date!!!

    // --------------------------------------------------------------
    // User Opcodes
    // --------------------------------------------------------------

    // 02004000  0x80 * 4  User Opcode Vector Table
    unsigned int UserOpcodeVectors[0x80];

    // 02004???         Padding to 02005000
    unsigned char Reserved5[0x1000 - 0x80 * 4];

    // --------------------------------------------------------------
    // VIDEO
    // --------------------------------------------------------------

    // 02005000    4    Flush Video Buffer Flag
    unsigned int VideoFlushBuffer;

    // 02005004    4    Hardware (Host System) Cursor Hidden (1 == hidden)
    unsigned int CursorHidden;

    // 02005008         Padding to 02006000
    unsigned char Reserved6[0x1000 - 2 * 4];  // XXX Keep this up to date!!!

    // 02006000  640*480*4    Flush Video Buffer Flag
    unsigned int VideoBuffer[640*480];

  };
} Special;

unsigned char Hdd[0x1000000];

union QueueEvent
{
  unsigned int type;
  struct EventMouseMove
  {
    unsigned int type;
    unsigned int x;
    unsigned int y;
  } MouseMove;

  struct EventButtonChange
  {
    unsigned int type;
    unsigned int x;
    unsigned int y;
    unsigned int LMB;
    unsigned int RMB;
    unsigned int MMB;
  } MouseButton;

  struct Keyboard
  {
    unsigned int type;
    unsigned int key;
  } Keyb;
};

enum EventType
{
  EVENT_NONE,
  EVENT_TIMER1,
  EVENT_TIMER2,
  EVENT_MOUSEMOVE,
  EVENT_MOUSEBUTTON,
  EVENT_KEYDOWN,
  EVENT_KEYUP
};

#define EVENT_QUEUE_SIZE 128
union QueueEvent EventQueue[EVENT_QUEUE_SIZE];
unsigned int EventQueueEnd;
unsigned int EventQueueStart; // End < Start

bool WasHDDChanged;

unsigned int CurrentTime;

unsigned int Timer1EndTime;
bool Timer1Online;

//#define DEBUGMODE 1

bool ROMWriteExceptionPending;
bool InvAddrExceptionPending;
bool ExceptionPending;
bool VMHalt;
SDL_Surface *MainScreen;

unsigned int ReleaseCounter;

union QueueEvent *PushEvent()
{
  // Space left ?
  if(EventQueueStart - EventQueueEnd >= EVENT_QUEUE_SIZE - 1)
    return NULL; // Event Queue full

  return &EventQueue[(EventQueueStart++) % EVENT_QUEUE_SIZE];
}

union QueueEvent *PopEvent()
{
  // Space left ?
  if(EventQueueStart == EventQueueEnd)
    return NULL; // Event Queue empty

  return &EventQueue[(EventQueueEnd++) % EVENT_QUEUE_SIZE];
}

bool FileSetContent(const char *filename, char *data, int sz)
{
  FILE *f;
  f = fopen(filename, "wb");
  if(f)
  {
    fwrite(data, 1, sz, f);
    fclose(f);
  }
  else
  {
    return false;
  }

  return true;
}

inline unsigned int InstGetDword(unsigned char *data)
{
  return *(unsigned int *)data;
}

unsigned int GetDword(unsigned int addr)
{
  if(ExceptionPending) return 0;

  unsigned int type = addr >> 24;
  unsigned int raddr = addr & 0x00ffffff;

  if(raddr > 0x00fffffc || type > 3)
  {
    ExceptionPending = true;
    InvAddrExceptionPending = true;
    return 0;
  }

  switch(type)
  {
    case 0: return *(unsigned int*)&Rom[raddr];
    case 1: return *(unsigned int*)&Ram[raddr];
    case 2: return *(unsigned int*)&Special.Data[raddr];
    case 3: return *(unsigned int*)&Hdd[raddr];
  }

  return 0;
}

void HandleSpecial(unsigned int addr, unsigned int data)
{
  if(addr == 0x0200100C)
  {
    // Timer1 start ?
    if(data == 0)
    {
      Timer1Online = false;
    }
    else
    {
      Timer1Online = true;
      Timer1EndTime = CurrentTime + data;
    }
  }
  else if(addr == 0x02005004)
  {
    // Hardware cursor change ?
#ifdef DEBUGMODE
    printf("Changing cursor hide state to %s\n", data ? "HIDE" : "SHOW");
#endif
    SDL_ShowCursor(!data);
  }
}

void SetDword(unsigned int addr, unsigned int data)
{
  if(ExceptionPending) return;

  unsigned int type = addr >> 24;
  unsigned int raddr = addr & 0x00ffffff;

  if(raddr > 0x00fffffc || type > 3)
  {
    ExceptionPending = true;
    InvAddrExceptionPending = true;
    return;
  }

  //printf("raddr == %.8x\n", raddr); fflush(stdout);

  switch(type)
  {
    case 0:
      ExceptionPending = true;
      ROMWriteExceptionPending = true;
      break;

    case 1: *(unsigned int*)&Ram[raddr] = data; break;
    case 2: *(unsigned int*)&Special.Data[raddr] = data; 
            HandleSpecial(addr, data);
            break;
    case 3: WasHDDChanged = true; 
            *(unsigned int*)&Hdd[raddr] = data; break;
  }
}

unsigned char* GetAddr(unsigned int addr)
{
  if(ExceptionPending) return NULL;

  unsigned int type = addr >> 24;
  unsigned int raddr = addr & 0x00ffffff;

  if(raddr > 0x00fffffc || type > 3)
  {
    ExceptionPending = true;
    InvAddrExceptionPending = true;
    return NULL;
  }

  switch(type)
  {
    case 0: return &Rom[raddr];
    case 1: return &Ram[raddr];
    case 2: return &Special.Data[raddr];
    case 3: return &Hdd[raddr];
  }

  return NULL; // this will never happen
}

void CrashDump(void)
{
  printf("Crashdumping RAM...\n");
  FileSetContent("CrashDump.RAM", (char*)Ram, sizeof(Ram));

  printf("Crashdumping Special...\n");
  FileSetContent("CrashDump.Spec", (char*)Special.Data, sizeof(Special.Data));
}

int
main(int argc, char **argv)
{
  FILE *f;

  // Init SDL
  if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE ) !=0)
  {
    return 2;
  }

  MainScreen = SDL_SetVideoMode(640, 480, 32 ,SDL_HWSURFACE | SDL_DOUBLEBUF);
  SDL_WM_SetCaption("Hardcode C0MP0 Virtual Machine", 0);

  // Read ROM
  if(argc < 2)
  {
    printf("Usage: vm <ROM>\n");
    return 1;
  }

  f = fopen(argv[1], "rb");
  if(!f)
  {
    printf("ROM not found\n");
    return 2;
  }

  fread(Rom, 1, sizeof(Rom), f);
  fclose(f);

  // Read HDD
  f = fopen("HDD.bin", "rb");
  if(f)
  {
    printf("Reading HDD.bin... %i bytes read!\n",
           fread(Hdd, 1, sizeof(Hdd), f));

    fclose(f);
  }
  else
  {
    puts("HDD not found. It will be created.");
  }

  unsigned int ExecTime = SDL_GetTicks();
  unsigned int HddTime = ExecTime + 5000; // 5 Sec

#ifdef BENCHMARK
  unsigned int Bench_FrameCount = 0;
  unsigned int Bench_NextTime = ExecTime + 1000;
#endif

  // Execute
  while(!VMHalt)
  {
    // Current time
    CurrentTime = SDL_GetTicks();

#ifdef BENCHMARK
    if(CurrentTime >= Bench_NextTime)
    {
      printf("VMCPU cycles per second: %u\n", Bench_FrameCount);
      Bench_NextTime = CurrentTime + 1000;
      Bench_FrameCount = 0;
    }
    Bench_FrameCount++;
#endif

    // Setup
    ROMWriteExceptionPending = false;
    InvAddrExceptionPending  = false;
    ExceptionPending         = false;

    // Release Blocker ?
    if(ReleaseCounter)
    {
      if(--ReleaseCounter == 0)
      {
        // Release!
        if(Special.IntBlocker > 0)
        {
          Special.IntBlocker--;
        }
#ifdef DEBUGMODE
        else
        {
          puts("Trying to Release IntBlocker when it is not held.");
        }
#endif        

      }
    }

    // Event queue
    if(!Special.IntBlocker)
    {
      union QueueEvent *qev = PopEvent();
      if(qev)
      {
#ifdef DEBUGMODE
        puts("Handling queued event.");
#endif        
        switch(qev->type)
        {
          case EVENT_TIMER1:
            if(Special.Timer1Handler)
            {
              // Go to timer
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.Timer1Handler;
#ifdef DEBUGMODE
              puts("Queue: Timer1");
#endif
            }
            else
            {
              puts("Queue: Timer1 hit, but no handler???");
            }
            break;

          case EVENT_TIMER2:
            if(Special.Timer2Handler)
            {
              // Go to timer
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.Timer2Handler;
#ifdef DEBUGMODE
              puts("Queue: Timer2");
#endif
            }
            else
            {
              puts("Queue: Timer2 hit, but no handler???");
            }
            break;

          case EVENT_MOUSEMOVE:
            if(Special.MouseMoveHandler)
            {
              // Go to timer
              Special.MouseX = qev->MouseMove.x;
              Special.MouseY = qev->MouseMove.y;
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.MouseMoveHandler;
#ifdef DEBUGMODE
              puts("Queue: MouseMove");
#endif
            }
            else
            {
              puts("Queue: MouseMove hit, but no handler???");
            }
            break;

          case EVENT_MOUSEBUTTON:
            if(Special.MouseButtonChangeStateHandler)
            {
              // Go to timer
              Special.MouseX = qev->MouseButton.x;
              Special.MouseY = qev->MouseButton.y;
              Special.MouseLMBState = qev->MouseButton.LMB;
              Special.MouseRMBState = qev->MouseButton.RMB;
              Special.MouseMMBState = qev->MouseButton.MMB;
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.MouseButtonChangeStateHandler;
#ifdef DEBUGMODE
              puts("Queue: MouseButton");
#endif
            }
            else
            {
              puts("Queue: MouseButton hit, but no handler???");
            }
            break;

          case EVENT_KEYDOWN:
            if(Special.KeyDownHandler)
            {
              // Go to timer
              Special.KeyHandled = qev->Keyb.key;
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.KeyDownHandler;
#ifdef DEBUGMODE
              puts("Queue: KeyDown");
#endif
            }
            else
            {
              puts("Queue: KeyDown hit, but no handler???");
            }
            break;

          case EVENT_KEYUP:
            if(Special.KeyUpHandler)
            {
              // Go to timer
              Special.KeyHandled = qev->Keyb.key;
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.KeyUpHandler;
#ifdef DEBUGMODE
              puts("Queue: KeyUp");
#endif
            }
            else
            {
              puts("Queue: KeyUp hit, but no handler???");
            }
            break;

          default: break;
        }
      }
    }

    // Timers
    Special.TimeMS = CurrentTime - ExecTime;

    if(Timer1Online && Timer1EndTime <= CurrentTime && Special.Timer1Handler)
    {
      // Turn off
      Timer1EndTime = 0;
      Timer1Online  = false;

      // Go
      if(Special.IntBlocker)
      {
#ifdef DEBUGMODE
        puts("Timer2 hit, but blocked. Adding to queue.");
#endif
        union QueueEvent *qev = PushEvent();
        if(!qev)
        {
#ifdef DEBUGMODE
          puts("Queue full. Can't queue event.");
#endif
        }
        else
        {
          qev->type = EVENT_TIMER1;
        }
      }
      else
      {
        // Go to timer
        Special.IntBlocker++;
        Special.IntReturnAddr = Special.IP;
        Special.IP = Special.Timer1Handler;
#ifdef DEBUGMODE
        printf("Timer1 hit! Timer1 off...\n");
#endif
      }
    }

    if(Special.Timer2IntFreq && Special.Timer2Handler)
    {
      Special.Timer2IntFreq--;

      if(Special.Timer2IntFreq == 0)
      {
        // Go
        if(Special.IntBlocker)
        {
#ifdef DEBUGMODE
          puts("Timer2 hit, but blocked. Adding to queue.");
#endif

          union QueueEvent *qev = PushEvent();
          if(!qev)
          {
#ifdef DEBUGMODE
            puts("Queue full. Can't queue event.");
#endif
          }
          else
          {
            qev->type = EVENT_TIMER2;
          }
        }
        else
        {
          // Go to timer
          Special.IntBlocker++;
          Special.IntReturnAddr = Special.IP;
          Special.IP = Special.Timer2Handler;
#ifdef DEBUGMODE
          printf("Timer2 hit! Timer2 off...\n");
#endif
        }
      }
    }    

    // Inc Counter
    Special.ExecInstCounterLow++;
    if(Special.ExecInstCounterLow == 0)
      Special.ExecInstCounterHi++;

    // HDD save
    if(WasHDDChanged && HddTime < CurrentTime)
    {
      WasHDDChanged = 0;

#ifdef DEBUGMODE
      printf("Saving HDD..."); fflush(stdout);
#endif
      FileSetContent("HDD.bin", (char*)Hdd, sizeof(Hdd));
#ifdef DEBUGMODE
      printf("Done!\n"); fflush(stdout);
#endif

      HddTime = CurrentTime + 5000; // 5 Sec
    }

    // Let's start with the events
    SDL_Event ev;
    while(SDL_PollEvent(&ev))
    {
      switch(ev.type)
      {
        // MOUSE MOVE
        case SDL_MOUSEMOTION:
          // Move mouse
          if(!Special.IntBlocker)
          {
            Special.MouseX = ev.motion.x;
            Special.MouseY = ev.motion.y;
          }

          // Check if handler
          if(Special.MouseMoveHandler)
          {
            if(Special.IntBlocker && !Special.DisableMouseMoveQueue)
            {
#ifdef DEBUGMODE
              puts("Event MouseMove, handled but blocked (queued)");
#endif
              union QueueEvent *qev = PushEvent();
              if(!qev)
              {
#ifdef DEBUGMODE
                puts("Queue full. Can't queue event.");
#endif
                break;
              }

              qev->type = EVENT_MOUSEMOVE;
              qev->MouseMove.x = ev.motion.x;
              qev->MouseMove.y = ev.motion.y;
            }
            else
            {
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.MouseMoveHandler;
            }
          }
          break;

        // MOUSE BUTTON CHANGE
        case SDL_MOUSEBUTTONDOWN:
        case SDL_MOUSEBUTTONUP:

            if(ev.button.button == SDL_BUTTON_LEFT)
              Special.MouseLMBState = ev.button.state == SDL_PRESSED ? 1 : 0;
            else if(ev.button.button == SDL_BUTTON_RIGHT)
              Special.MouseRMBState = ev.button.state == SDL_PRESSED ? 1 : 0;
            else if(ev.button.button == SDL_BUTTON_MIDDLE)
              Special.MouseMMBState = ev.button.state == SDL_PRESSED ? 1 : 0;

          // Check if handler
          if(Special.MouseButtonChangeStateHandler)
          {
            if(Special.IntBlocker)
            {
#ifdef DEBUGMODE
              puts("Event MouseButtonChangeState, handled but blocked (queued)");
#endif
              union QueueEvent *qev = PushEvent();
              if(!qev)
              {
#ifdef DEBUGMODE
                puts("Queue full. Can't queue event.");
#endif
                break;
              }

              qev->type = EVENT_MOUSEBUTTON;
              qev->MouseButton.x = ev.motion.x;
              qev->MouseButton.y = ev.motion.y;
              qev->MouseButton.LMB = Special.MouseLMBState;
              qev->MouseButton.RMB = Special.MouseRMBState;
              qev->MouseButton.MMB = Special.MouseMMBState;

              if(ev.button.button == SDL_BUTTON_LEFT)
                qev->MouseButton.LMB = ev.button.state == SDL_PRESSED ? 1 : 0;
              else if(ev.button.button == SDL_BUTTON_RIGHT)
                qev->MouseButton.RMB = ev.button.state == SDL_PRESSED ? 1 : 0;
              else if(ev.button.button == SDL_BUTTON_MIDDLE)
                qev->MouseButton.MMB = ev.button.state == SDL_PRESSED ? 1 : 0;
            }
            else
            {
#ifdef DEBUGMODE
            puts("Event MouseButtonChangeState, handled");
#endif
              // Set this to be OK
              Special.MouseX = ev.button.x;
              Special.MouseY = ev.button.y;

              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.MouseButtonChangeStateHandler;
            }
          }
          else
          {
#ifdef DEBUGMODE
            puts("Event MouseButtonChangeState, not handled");
#endif
          }
          break;

        // KEY DOWN
        case SDL_KEYDOWN:
          // Mark as down
          Special.KeyMap[ev.key.keysym.sym] = 1;

          // Check handler
          if(Special.KeyDownHandler)
          {
            if(Special.IntBlocker)
            {
#ifdef DEBUGMODE
              puts("Event KeyDown, handled but blocked (queued)");
#endif

              union QueueEvent *qev = PushEvent();
              if(!qev)
              {
#ifdef DEBUGMODE
                puts("Queue full. Can't queue event.");
#endif
                break;
              }

              qev->type = EVENT_KEYDOWN;
              qev->Keyb.key = ev.key.keysym.sym;
            }
            else
            {
#ifdef DEBUGMODE
              puts("Event KeyDown, handled");
#endif
              Special.KeyHandled = ev.key.keysym.sym;
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.KeyDownHandler;
            }
          }
#ifdef DEBUGMODE
          else
          {
            puts("Event KeyDown, but no handler");
          }
#endif
          break;

        // KEY UP
        case SDL_KEYUP:
          // Mark as up
          Special.KeyMap[ev.key.keysym.sym] = 0;

          // Check handler
          if(Special.KeyUpHandler)
          {
            if(Special.IntBlocker)
            {
#ifdef DEBUGMODE
              puts("Event KeyUp, handled but blocked (queued)");
#endif

              union QueueEvent *qev = PushEvent();
              if(!qev)
              {
#ifdef DEBUGMODE
                puts("Queue full. Can't queue event.");
#endif
                break;
              }

              qev->type = EVENT_KEYUP;
              qev->Keyb.key = ev.key.keysym.sym;

            }
            else
            {
#ifdef DEBUGMODE
              puts("Event KeyUp, handled");
#endif
              Special.KeyHandled = ev.key.keysym.sym;
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.KeyUpHandler;
            }
          }          
#ifdef DEBUGMODE
          else
          {
            puts("Event KeyUp, but no handler");
          }
#endif
          break;

        case SDL_QUIT:
            VMHalt = true;
          break;
      }
    }

    unsigned char *ptr = GetAddr(Special.IP);

    // Got anything ?
    if(!ptr)
    {
      // HUH!
      // GOGO! Make an exception they will NEVER forget!
      if(Special.ExceptionInvAddrHandler)
      {
        if(Special.IntBlocker)
        {
          // Well, what to say... Die()
          printf("Blocked Special.IP Inv Addr Exception: Void-Opcode at %.8x\n", Special.IP);
          VMHalt = true;
          CrashDump();
          break;
        }

#ifdef DEBUGMODE
        printf("Handled Special.IP Inv Addr Exception: Void-Opcode at %.8x\n", Special.IP);
#endif

        // OK, Go to handler
        Special.IntBlocker++;
        Special.IntReturnAddr = Special.IP;
        Special.IP = Special.ExceptionInvAddrHandler;
        continue;
      }
      else
      {
        // Well, what to say... Die()
        printf("Unhandled Special.IP Inv Addr Exception: Void-Opcode at %.8x\n", Special.IP);
        VMHalt = true;
        CrashDump();
        break;
      }
    }

    InstructionFunc iptr = FuncPtrs[*ptr];

//    printf("Executing %.2x at %.8x\n", *ptr, Special.IP); fflush(stdout);

    if(*ptr < 0x80) // Built-in opcodes
    {
      if(iptr)
      {
        // Execute
        iptr(ptr + 1);

        // Check if there is an exception
        if(ExceptionPending)
        {
          // HUH! Someone messed up good!
          if(ROMWriteExceptionPending)
          {
            // Can be handled ?
            if(Special.ExceptionROMWriteErrorHandler)
            {
              if(Special.IntBlocker)
              {
                // Well, what to say... Die()
                printf("Blocked ROM Write Exception: Opcode (%.2x) at %.8x \n", *ptr, Special.IP);
                VMHalt = true;
                CrashDump();
                break;
              }

#ifdef DEBUGMODE
              printf("Handled ROM Write Exception: Opcode (%.2x) at %.8x \n", *ptr, Special.IP);
#endif

              // OK, Go to handler
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.ExceptionROMWriteErrorHandler;
              continue;
            }
            else
            {
              // Well, what to say... Die()
              printf("Unhandled ROM Write Exception: Opcode (%.2x) at %.8x \n", *ptr, Special.IP);
              VMHalt = true;
              CrashDump();
              break;
            }
          }

          if(InvAddrExceptionPending)
          {
            // Can be handled ?
            if(Special.ExceptionInvAddrHandler)
            {
              if(Special.IntBlocker)
              {
                // Well, what to say... Die()
                printf("Blocked Inv Addr Exception: Opcode (%.2x) at %.8x \n", *ptr, Special.IP);
                VMHalt = true;
                CrashDump();
                break;
              }

#ifdef DEBUGMODE
              printf("Handled Inv Addr Exception: Opcode (%.2x) at %.8x \n", *ptr, Special.IP);
#endif

              // OK, Go to handler
              Special.IntBlocker++;
              Special.IntReturnAddr = Special.IP;
              Special.IP = Special.ExceptionInvAddrHandler;
              continue;
            }
            else
            {
              // Well, what to say... Die()
              printf("Unhandled Inv Addr Exception: Opcode (%.2x) at %.8x \n", *ptr, Special.IP);
              VMHalt = true;
              CrashDump();
              break;
            }
          }


        }        
      }
    }
    else // User opcodes
    {
      unsigned char user_opcode = *ptr - 0x80;

      // Check if handled
      if(Special.UserOpcodeVectors[user_opcode])
      {
        // Check if blocked
        /*if(Special.IntBlocker)
        {
          // Well, what to say... Die()
          printf("User Opcode Blocked: User Opcode (%.2x) at %.8x was blocked (Interrupt in progress). Terminating.\n", *ptr, Special.IP);
          VMHalt = true;
          CrashDump();
          break;
        }
        else
        {*/
          // All OK
          Special.IntBlocker++;
          Special.IntReturnAddr = Special.IP;
          Special.IP = Special.UserOpcodeVectors[user_opcode];
          continue;
        /*}*/

      }
      else
      {
        // Issue exception
        iptr = NULL; // This will be handled later
      }
    }

    // Exception ?
    if(iptr == NULL)
    {
      if(Special.ExceptionInvalidOpcodeHandler)
      {
        // Check if interrupt in progress 
        if(Special.IntBlocker)
        {
          // Uh, this can't be queued... Die()
          printf("Exception Blocked: Invalid Opcode (%.2x) at %.8x\n", *ptr, Special.IP);
          CrashDump();
          VMHalt = true;
          break;
        }      
        else
        {
#ifdef DEBUGMODE
          printf("Handled Exception: Invalid Opcode (%.2x) at %.8x\n", *ptr, Special.IP);
#endif
          // Handle exception
          Special.IntBlocker++;
          Special.IntReturnAddr = Special.IP;
          Special.IP = Special.ExceptionInvalidOpcodeHandler;
          continue;
        }
      }
      else
      {
        printf("Unhandled Exception: Invalid Opcode (%.2x) at %.8x\n", *ptr, Special.IP);
        VMHalt = true;
        CrashDump();
        break;
      }
    }

    // Redraw Screen ?
    if(Special.VideoFlushBuffer)
    {
     // puts("Flushing video buffer");
      Special.VideoFlushBuffer = 0;

      // TODO: Opt this!
      memcpy(MainScreen->pixels, Special.VideoBuffer, sizeof(Special.VideoBuffer));      

      // Flip
      SDL_Flip(MainScreen);
    }
  }

  // Save HDD
  FileSetContent("HDD.bin", (char*)Hdd, sizeof(Hdd));

  SDL_FreeSurface(MainScreen);
  SDL_Quit();

  return 0;
}

InstructionFuncType(Halt)
{
  (void)data;
  VMHalt = true;
}

InstructionFuncType(Move_MemImm)
{
  SetDword(InstGetDword(data), InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Move_MemMem)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data+4)));

  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Add_MemImm)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) + InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Add_MemMem)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) + GetDword(InstGetDword(data+4)));

  Special.IP += 1 + 4 + 4;
}


InstructionFuncType(And_MemImm)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) & InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(And_MemMem)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) & GetDword(InstGetDword(data+4)));

  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Or_MemImm)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) | InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Or_MemMem)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) | GetDword(InstGetDword(data+4)));

  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Xor_MemImm)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) ^ InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Shl_MemImm)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) << InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Shr_MemImm)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) >> InstGetDword(data+4));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Xor_MemMem)
{
  SetDword(InstGetDword(data), GetDword(InstGetDword(data)) ^ GetDword(InstGetDword(data+4)));

  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(Goto_Mem)
{
  Special.IP = GetDword(InstGetDword(data));
}

InstructionFuncType(Goto_Imm)
{
  Special.IP = InstGetDword(data);
}

InstructionFuncType(Neg_Mem)
{
  SetDword(InstGetDword(data), ~GetDword(InstGetDword(data)));

  Special.IP += 1 + 4;
}

InstructionFuncType(MovePtrFrom_MemMem)
{
  SetDword(InstGetDword(data), GetDword(GetDword(InstGetDword(data+4))));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(MovePtrTo_MemMem)
{
  SetDword(GetDword(InstGetDword(data)), GetDword(InstGetDword(data+4)));
  
  Special.IP += 1 + 4 + 4;
}

InstructionFuncType(GotoEqual_ImmMemMem)
{
  if(GetDword(InstGetDword(data+4)) == GetDword(InstGetDword(data+8)))
  {
    Special.IP = InstGetDword(data);
    return;
  }

  Special.IP += 1 + 4 + 4 + 4;
}

InstructionFuncType(GotoLess_ImmMemMem)
{
  if(GetDword(InstGetDword(data+4)) < GetDword(InstGetDword(data+8)))
  {
    Special.IP = InstGetDword(data);
    return;
  }

  Special.IP += 1 + 4 + 4 + 4;
}

InstructionFuncType(BlockInt)
{
  (void)data;
  Special.IntBlocker++;
  Special.IP++;
}

InstructionFuncType(ReleaseInt)
{
  (void)data;
  //Special.IntBlocker = 0;
  ReleaseCounter = 2;
  Special.IP++;
}

inline bool SIMDCheckSrc(unsigned int Src, unsigned int Count)
{
  if((Src >> 24) != ((Src + Count) >> 24) ||
    (Src >> 24) > 3)
  {
    ExceptionPending = true;
    InvAddrExceptionPending = true;
    return false;
  }

  return true;
}

inline bool SIMDCheckDst(unsigned int Dst, unsigned int Count)
{
  if((Dst >> 24) != ((Dst + Count) >> 24) ||
     (Dst >> 24) > 3)
  {
    ExceptionPending        = true;
    InvAddrExceptionPending = true;
    return false;
  }

  /* This is checked later
  if((Dst >> 24) == 0)
  {
    ExceptionPending         = true;
    ROMWriteExceptionPending = true;
    return false;
  }
  */

  return true;
}

#define DWORD unsigned int

/* 16 */ InstructionFuncType(SIMDMove_MemMemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Src = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8)) * 4;

  if(ExceptionPending) return;
  if(!SIMDCheckSrc(Src, Count)) return;
  if(!SIMDCheckDst(Dst, Count)) return;

  unsigned char *pDst = GetAddr(Dst);
  unsigned char *pSrc = GetAddr(Src);
  if(!pDst || !pSrc) return;

  memmove(pDst, pSrc, Count);

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 17 */ InstructionFuncType(SIMDAdd_MemMemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Src = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckSrc(Src, Count * 4)) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  DWORD *pSrc = (DWORD*)GetAddr(Src);
  if(!pDst || !pSrc) return;

  for(i = 0; i < Count; i++, pDst++, pSrc++)
    *pDst += *pSrc;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 18 */ InstructionFuncType(SIMDAnd_MemMemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Src = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckSrc(Src, Count * 4)) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  DWORD *pSrc = (DWORD*)GetAddr(Src);
  if(!pDst || !pSrc) return;

  for(i = 0; i < Count; i++, pDst++, pSrc++)
    *pDst &= *pSrc;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 19 */ InstructionFuncType(SIMDOr_MemMemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Src = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckSrc(Src, Count * 4)) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  DWORD *pSrc = (DWORD*)GetAddr(Src);
  if(!pDst || !pSrc) return;

  for(i = 0; i < Count; i++, pDst++, pSrc++)
    *pDst |= *pSrc;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 1A */ InstructionFuncType(SIMDXor_MemMemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Src = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckSrc(Src, Count * 4)) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  DWORD *pSrc = (DWORD*)GetAddr(Src);
  if(!pDst || !pSrc) return;

  for(i = 0; i < Count; i++, pDst++, pSrc++)
    *pDst ^= *pSrc;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 1B */ InstructionFuncType(SIMDShl_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst <<= Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 1C */ InstructionFuncType(SIMDShr_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst >>= Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 1D */ InstructionFuncType(SIMDNeg_MemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Count = GetDword(InstGetDword(data+4));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst = ~(*pDst);

  Special.IP += 1 + 4 + 4;
}


/* 1E */ InstructionFuncType(SIMDCopy_MemMemMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Src = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8)) * 4;

  if(ExceptionPending) return;
  if(!SIMDCheckSrc(Src, Count)) return;
  if(!SIMDCheckDst(Dst, Count)) return;

  unsigned char *pDst = GetAddr(Dst);
  unsigned char *pSrc = GetAddr(Src);
  if(!pDst || !pSrc) return;

  memcpy(pDst, pSrc, Count);

  Special.IP += 1 + 4 + 4 + 4;    
}

// SESESE2
/* 1F */ InstructionFuncType(SIMDAdd_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst += Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 20 */ InstructionFuncType(SIMDAnd_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst &= Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 21 */ InstructionFuncType(SIMDOr_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst |= Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 22 */ InstructionFuncType(SIMDXor_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst ^= Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

/* 23 */ InstructionFuncType(SIMDSet_MemImmMem)
{
  DWORD Dst = InstGetDword(data);
  DWORD Imm = InstGetDword(data+4);
  DWORD Count = GetDword(InstGetDword(data+8));
  DWORD i;

  if(ExceptionPending) return;
  if(!SIMDCheckDst(Dst, Count * 4)) return;

  DWORD *pDst = (DWORD*)GetAddr(Dst);
  if(!pDst) return;

  for(i = 0; i < Count; i++, pDst++)
    *pDst = Imm;

  Special.IP += 1 + 4 + 4 + 4;    
}

#undef DWORD

